"""
:requires: torch for compare_images_multi_sets_squeezed_test()
:sub_package: open_cv with the test
"""
try:
    from wizzi_utils.pyplot.test.test_pyplot_tools import *
except ModuleNotFoundError as e:
    pass
