"""
:requires: matplotlib, mpl_toolkits,
    torch for compare_images_multi_sets_squeezed()
"""
try:
    from wizzi_utils.pyplot.pyplot_tools import *
except ModuleNotFoundError as e:
    pass

from wizzi_utils.pyplot import test
